import unittest
from hein_control.mixin.reg import InstanceRegistry
from hein_control.action.basic import Action


class Extra(InstanceRegistry):
    """extra class to validate segregation"""
    pass


def test():
    return True


xtra = Extra()
action = Action(test)


class TestRegistry(unittest.TestCase):
    def test_double_registry(self):
        """ensures double registry of actions is not allowed
        todo move this to test_actions
        """
        self.assertRaises(
            ValueError,
            Action,
            test
        )

    def test_retrieval(self):
        """test general retrieval methods"""
        # retrieve decision instance
        first = Action.registered_actions[0]

        # attempt various ways of retrieving instance
        Action.class_instance_by_id(first._uuid)
        InstanceRegistry.get_instance_by_uuid(first._uuid)
        InstanceRegistry.get_instance_by_uuid(first.uuid)

    def test_cross_retrieval(self):
        """tests for retrieval by uuid from other classes"""
        first = Action.class_registry[0]
        Extra.search_class_for_id(
            first._uuid,
            'Action'
        )

    def test_segregation(self):
        """tests that registries are separated"""
        actions = Action.registered_actions
        self.assertFalse(
            all([isinstance(action, Extra) for action in actions])
        )



import copy
from functools import wraps
from typing import Callable, Iterable


def _update_args(original_args: tuple,
                 new_args: tuple,
                 ) -> tuple:
    """
    Updates the original argument tuple with the new arguments.

    :param original_args: original argument tuple
    :param new_args: new argument tuple
    :return: updated argument tuple
    """
    original_args = list(original_args)
    for ind, val in enumerate(new_args):
        try:
            original_args[ind] = val
        except IndexError:
            original_args.append(val)
    return tuple(original_args)


def is_builtin_type(obj):
    """returns whether the object is a builtin type"""
    return obj.__class__.__module__ == 'builtins'


def _update_kwargs(original_kwargs: dict,
                   new_kwargs: dict,
                   ) -> dict:
    """
    Updates the original kwargs with the new kwargs. Prevents mutation of the original kwargs.

    Only perform deepcopy on builtin types and interface instances that have specified the deepcopy method.

    :param original_kwargs: original keyword arguments
    :param new_kwargs: new keyword arguments
    :return: consolidated and updated keyword arguments
    """
    # todo compare provided types with signature types
    # create deepcopies of supporting types
    dct = {
        key: copy.deepcopy(value)
        for key, value in original_kwargs.items()
        if hasattr(value, '__deepcopy__') or is_builtin_type(value)
    }
    # add direct references for non-copyable types
    dct.update({
        key: original_kwargs[key] for key in original_kwargs.keys() - dct.keys()
    })
    # finally update with new kwargs
    dct.update(new_kwargs)
    return dct


class Configuration:
    def __init__(self,
                 *args,
                 **kwargs,
                 ):
        """
        Class for storing a configuration for a function execution.

        :param args: arguments
        :param kwargs: keyword arguments
        """
        self._freeze = False
        self._args = args
        self._kwargs = kwargs
        # todo
        #   - update config from ConfiguredAction
        #   - consider moving to basic and leveraging this for Action
        #   - create uuid for each configuration, link configured action to configuration by uuid

    @property
    def _kwarg_string(self) -> str:
        """kwarg representation string"""
        return ", ".join([f"{key}={val}" for key, val in self._kwargs.items()])

    @property
    def _arg_string(self) -> str:
        """arg representation string"""
        return ", ".join([str(arg) for arg in self._args])

    def __repr__(self):
        strings = []
        args = self._arg_string
        if args != '':
            strings.append(args)
        kw = self._kwarg_string
        if kw != '':
            strings.append(kw)
        return f'{self.__class__.__name__}({", ".join(strings)})'

    def __str__(self):
        return self.__repr__()

    def __copy__(self):
        return self.__class__(
            *self._args,
            **self._kwargs,
        )

    def _ensure_unfrozen(fn: Callable):
        """decorator which will do the thing you want"""
        @wraps(fn)
        def decorated(self: 'Configuration', *args, **kwargs):
            if self._freeze is True:
                raise ValueError(f'the instance is frozen and setting or updating of the args or kwargs is disabled')
            return fn(self, *args, **kwargs)
        return decorated

    @_ensure_unfrozen
    def update_args(self, *args):
        """Updates the arguments for the Action"""
        self.args = _update_args(
            self._args,
            args
        )

    @_ensure_unfrozen
    def update_kwargs(self, **kwargs):
        """updates the keyword arguments for the Action"""
        self.kwargs = _update_kwargs(
            self._kwargs,
            kwargs
        )

    @property
    def args(self) -> tuple:
        """arguments for the action"""
        return self._args

    @args.setter
    @_ensure_unfrozen
    def args(self, value: Iterable):
        self._args = value

    @property
    def kwargs(self) -> dict:
        """keyword arguments for the action"""
        return self._kwargs

    @kwargs.setter
    @_ensure_unfrozen
    def kwargs(self, value: dict):
        self._kwargs = value

    @property
    def freeze(self) -> bool:
        """freezes setting or updating of args and kwargs for the instance"""
        return self._freeze

    @freeze.setter
    def freeze(self, value: bool):
        self._freeze = value

    _ensure_unfrozen = staticmethod(_ensure_unfrozen)


"""
Tools for scheduling sequences of actions to execute at prescribed times.
"""
import os
import threading
import time
import logging
import bisect
import datetime
from typing import List, Union, Iterable
from .timepoint import ActionTimePoint
from .action import Action, ConfiguredAction
from .sequencing import ConfiguredSequence
from .mixin import InstanceRegistry

logger = logging.getLogger(__name__)


class SamplingScheduler(InstanceRegistry):
    # pause time between cycle checks
    CYCLE_TIME = 0.5

    # default state for whether to wait for completion of previous trigger if busy
    WAIT_IF_BUSY = True

    UUID_PREFIX = 'sched-'

    def __init__(self,
                 *actions: Union[str, callable, Action, ConfiguredAction],
                 time_deltas: List[float] = None,
                 sequence_name: str = None,
                 ):
        """
        Scheduler for sequences of configured actions.

        :param time_deltas: a list of time deltas to initialize the sequence with (passed directly to
            insert_time_point_list)
        :param actions: callable action or the name of a callable action previously defined as a SchedulerAction
        :param sequence_name: convenience reference name for the current experiment
        """
        InstanceRegistry.__init__(self)
        logger.debug(f'creating scheduler {self._uuid[:5]}')
        self.logger = logger.getChild(f'{self.__class__.__name__} {self._uuid[:5]}')

        # event control of killswitch for stop command
        self._killswitch = threading.Event()

        # create sequence
        self.sequence: ConfiguredSequence = ConfiguredSequence(
            *actions,
            sequence_name=sequence_name,
        )

        # time stamp for Scheduler creation
        self.created = time.time()
        # start time for auto-stepping
        self._start_time: float = None
        # thread for the sequence
        self._sequence_thread: threading.Thread = threading.Thread(
            target=self._sequence_loop,
            name=f'{self.__class__.__name__} sequence thread',
            daemon=True,
        )
        self._sequence_pause: threading.Event = threading.Event()
        # dictionary for time points
        self._time_points: List[ActionTimePoint] = []

        if time_deltas is not None:
            self.insert_time_point_list(*time_deltas)

        # store experiment name for reference
        self.sequence_name: str = sequence_name

        # todo implement databasing
        #   - add method which reinstantiates from database
        # todo have the experiment be named based on the sample number in the sequence
        # todo implement ability to change action associated with time points
        # todo generalize action parsing
        # todo add run_forever method
        # todo add attribute that defines default behaviour when all timepoints are completed (auto kill/cleanup)
        # todo add restart method (restarts the whole thing)

    def __repr__(self):
        return (
            f'{self.__class__.__name__} {self._uuid[:5]} '
            f'{" STARTED" if self.started is True else ""} '
            f'{" TERMINATED" if self.terminated else ""}'
        )

    def __str__(self):
        return self.__repr__()

    @property
    def uuid(self) -> str:
        """uuid4 for the scheduler"""
        return self.UUID_PREFIX + self._uuid

    @property
    def started(self) -> bool:
        """whether the sequence has been started"""
        return self._sequence_thread.is_alive()

    @property
    def terminated(self) -> bool:
        """whether the sequence has been terminated"""
        return self._killswitch.is_set()

    @property
    def current_time(self) -> float:
        """the current time since starting the scheduler"""
        if self._start_time is not None:
            return time.time() - self._start_time

    @property
    def time_points(self) -> List[ActionTimePoint]:
        """list of time points for the sequence"""
        return self._time_points

    @property
    def triggered_time_points(self) -> int:
        """number of triggered time points"""
        return len([tp for tp in self.time_points if tp.triggered is True])

    @property
    def remaining_time_points(self) -> int:
        """number of time points remaining to be sampled"""
        return len([tp for tp in self.time_points if tp.triggered is False])

    @property
    def next_time_point(self) -> ActionTimePoint:
        """The next time point for triggering"""
        # todo filter out non-action points
        # filter to untriggered time points
        unsampled = [tp for tp in self._time_points if tp.triggered is False]
        if len(unsampled) == 0:
            return
        # return first up
        return min(
            unsampled,
            key=lambda x: x.time_delta
        )

    @property
    def estimated_time_until_next_point(self) -> float:
        """estimated time in seconds until next time point"""
        return  # todo

    @property
    def last_triggered_time_point(self) -> ActionTimePoint:
        """The last time point which was triggered"""
        # list of triggered time points
        triggered = [tp for tp in self.time_points if tp.triggered is True]
        # if there are triggered points, return most recent one
        if len(triggered) > 0:
            return triggered[-1]

    @property
    def busy(self) -> bool:
        """whether there are any time points currently executing"""
        last_point = self.last_triggered_time_point
        if last_point is not None and last_point.in_progress is True:
            return True
        return False

    @classmethod
    def get_newest_instance(cls) -> "SamplingScheduler":
        """Gets the newest instance of the sampling scheduler"""
        try:
            return cls.class_registry[-1]
        except IndexError:
            raise IndexError(f'The {cls.__name__} has no instances. ')

    def _determine_actions(self,
                           additional_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                           specified_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                           ) -> ConfiguredSequence:
        """
        Determines the actions for a time point. If no actions are provided, the default actions of the
        instance are used. If additional actions are provided, those actions will be performed in addition to the
        default actions (actions will be performed in order after the default actions are complete). If
        specified_actions are provided, only those actions will be performed.

        :param additional_actions: actions to run in addition to the default actions of the scheduler instance.
            These actions will be run after the default actions.
        :param specified_actions: actions to run instead of the default actions of the instance
        :return: actions for a time point
        """
        # establish list of actions
        if additional_actions is not None:
            actions = self.sequence.duplicate_configuration()
            actions.action_sequence.extend(additional_actions)
        elif specified_actions is not None:
            actions = ConfiguredSequence(*specified_actions)
        else:
            actions = self.sequence.duplicate_configuration()
        return actions

    def trigger(self,
                additional_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                specified_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                wait: bool = None,
                **action_kwargs
                ) -> ActionTimePoint:
        """
        Manually triggers the sequence with the actions provided. If no actions are provided, the default actions of the
        instance are used. If additional actions are provided, those actions will be performed in addition to the
        default actions (actions will be performed in order after the default actions are complete). If
        specified_actions are provided, only those actions will be performed.

        :param additional_actions: actions to run in addition to the default actions of the scheduler instance.
            These actions will be run after the default actions.
        :param specified_actions: actions to run instead of the default actions of the instance
        :param wait: whether to wait for completion of the actions before returning
        :param action_kwargs: keyword arguments for the actions. Keyword arguments will be passed through to their
            respective methods. See ActionTimePoint.trigger for details on kwarg handling
        :return: created ActionTimePoint instance
        """
        if self.busy is True:
            self.logger.warning(
                'trigger was called but the previous time point is still being executed. waiting...'
            )
            self.wait_for_time_point_completion()

        action_sequence = self._determine_actions(
            additional_actions=additional_actions,
            specified_actions=specified_actions,
        )

        # create a manual time point
        manual_time_point = ActionTimePoint(
            time.time() - self._start_time if self._start_time is not None else 0.,
            configuration=action_sequence,
        )
        # store start time if set
        if self._start_time is not None:
            manual_time_point.sequence_time_started = self._start_time
        # insert time point into list
        bisect.insort(
            self._time_points,
            manual_time_point
        )
        manual_time_point.trigger(
            wait=wait,
            **action_kwargs,
        )
        return manual_time_point

    def _sequence_loop(self):
        """
        Sequence loop that watches time points and triggers actions at the appropriate times
        """
        while True:
            # if the pause flag has been set
            if self._sequence_pause.is_set() is False:
                # todo consider how to handle missed time points
                #   - continue loop and note if missed?
                #   - rapid-fire sample to catch up?
                self.logger.info(f'sequence is paused, waiting for start_sequence call...')
                self._sequence_pause.wait()
                self.logger.info(f'sequence has been resumed')
            if self._killswitch.is_set():  # if killswitch is thrown, break out of thread
                # todo consider a timeout argument
                while self.busy is True:  # waits for current actions to be complete (if active)
                    time.sleep(0.1)
                # todo perform cleanup actions (if any)
                break
            # get next time point
            next_timepoint = self.next_time_point
            # if there is no next time point, break out
            if next_timepoint is None:
                self.logger.info('all time points have been completed, sequence is pausing pending more time points')
                self._sequence_pause.clear()
            # if the timepoint has happened, trigger and store time
            elif time.time() >= self._start_time + next_timepoint.time_delta:
                self.logger.info(f'{next_timepoint} reached')
                if self.busy is True:
                    self.logger.warning(
                        'the next time point has been reached but the previous time point '
                        'is still being executed. waiting...'
                    )
                    self.wait_for_time_point_completion()
                next_timepoint.sequence_time_started = self._start_time
                next_timepoint.trigger()
            # otherwise wait predefined time and continue loop
            else:
                time.sleep(self.CYCLE_TIME)

    def wait_for_time_point_completion(self):
        """If a time point is being executed, this method will wait for it to be complete"""
        if self.busy is True:
            self.logger.info('waiting for completion of previous trigger')
            while self.busy is True:
                time.sleep(0.5)

    def insert_time_point(self,
                          time_delta: float = None,
                          position: int = -1,
                          relative_to: str = 'start',
                          shift_subsequent: bool = True,
                          number_of_points: int = 1,
                          additional_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                          specified_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                          ):
        """
        Inserts a time point into the list of time points.

        :param time_delta: time delta in seconds from the reference (relative_to)
        :param position: position in the time point list to insert the new time point (default is the end of of the list)
        :param relative_to: whether the specified time delta should be relative to the sequence time ('start') or
            the time point previous to the insertion point ('previous')
        :param shift_subsequent: whether to shift subsequent time points in the time point list by the time delta
        :param number_of_points: number of points with these specification to add
        """
        action_sequence = self._determine_actions(
            additional_actions=additional_actions,
            specified_actions=specified_actions,
        )

        # todo adjust insert_time_point to differentiate between sampling and standard time points
        # if end of list position is desired
        if position == -1:
            position = len(self._time_points)

        # determine offset based on specification
        if relative_to.lower() == 'previous':
            if len(self._time_points) == 0:  # special case where there are no previous points
                delta_offset = 0.
            elif position == 0:  # special case at the beginning of the list
                delta_offset = 0.
            else:
                delta_offset = self._time_points[position - 1].time_delta
        elif relative_to.lower() == 'start':
            delta_offset = 0.
        else:
            raise ValueError(f'The relative_to value "{relative_to}" is invalid. Choose from "start" or "previous".')

        # add time points
        self.logger.debug(f'Adding {number_of_points} timepoints at {time_delta} s spacing in position {position}.')
        for i in range(number_of_points):
            delta = (i + 1) * time_delta + delta_offset
            bisect.insort(
                self._time_points,
                ActionTimePoint(
                    delta,
                    configuration=action_sequence,
                    name=f'{delta} s',
                )
            )

        # if shifting of subsequent timepoints is desired
        if shift_subsequent is True:
            for subsequent_timepoint in self._time_points[i + 1 + position:]:
                subsequent_timepoint.time_delta += (i + 1) * time_delta

        # if there the thread is alive and paused (pending more time points), clear flag to resume
        if self._sequence_thread.is_alive() is True and self._sequence_pause.is_set() is False:
            self._sequence_pause.set()

    def insert_time_point_list(self,
                               *time_deltas: float,
                               position: int = -1,
                               relative_to: str = 'previous',
                               additional_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                               specified_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                               ):
        """
        Inserts a list of time points into the current list of time points.

        :param time_deltas: list of time deltas (seconds)
        :param position: position in the current list to insert the incoming time points
        :param relative_to: whether the specified time delta should be relative to the sequence time ('start') or
            the time point previous to the insertion point ('previous'; default)
        """
        if position == -1:
            position = len(self._time_points)
        for ind, time_delta in enumerate(time_deltas):
            self.insert_time_point(
                time_delta=time_delta,
                position=position + ind,
                relative_to=relative_to,
                additional_actions=additional_actions,
                specified_actions=specified_actions,
            )

    def insert_time_point_list_pairs(self,
                                     *num_delta_pairs,
                                     position: int = -1,
                                     relative_to: str = 'previous',
                                     additional_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                                     specified_actions: Iterable[Union[str, callable, ActionTimePoint]] = None,
                                     ):
        """
        Inserts a list of time points in number-of-points/time-delta form.

        e.g. 3, 10 will result in 3 time points with 10 second spacing. 3, 10, 5, 20 will results in 3 time points
        with 10 second spacing followed by 5 time points with 20 second spacing.

        Pairs may also be provided in number/time-delta tuples. e.g. (3, 10), (5, 20)

        :param num_delta_pairs: pairs of arguments in (number-of-points, time_delta) form
        :param position: position in the current list to insert the incoming time points
        :param relative_to: whether the specified time delta should be relative to the sequence time ('start') or
            the time point previous to the insertion point ('previous'; default)
        """
        if type(num_delta_pairs[0]) in [tuple, list]:
            pass
        elif len(num_delta_pairs) % 2 == 0:
            num_delta_pairs = [num_delta_pairs[i:i + 2] for i in range(0, len(num_delta_pairs), 2)]
        else:
            raise ValueError('The arguments for this method are expected to be either a tuple of num/time_delta pairs '
                             'or a list of com')
        expanded = []
        for num, time_delta in num_delta_pairs:
            expanded.extend(
                [time_delta] * num
                # [time_delta * (i + 1) for i in range(num)]
            )
        # todo fix functionality for adding points relative to the start
        #   current functionality adds n times the same time delta relative to the start
        self.insert_time_point_list(
            *expanded,
            position=position,
            relative_to=relative_to,
            additional_actions=additional_actions,
            specified_actions=specified_actions,
        )

    def start_sequence(self):
        """
        Starts the sequence loop and triggers time points as specified
        """
        # set the GO flag
        self._sequence_pause.set()
        # if thread has not been started, store start time and start thread
        if self._sequence_thread.is_alive() is False:
            self.logger.info('starting sequence thread')
            self._start_time = time.time()
            if any([tp.triggered for tp in self.time_points]):  # if any have already been triggered
                for tp in self.time_points:
                    if tp.triggered:
                        # todo figure out a better catch for time=0
                        # store the start time and calculate the effective time delta
                        tp.sequence_time_started = self._start_time
                        tp.time_delta = tp.relative_start_times[
                            min(tp.relative_start_times, key=lambda x: tp.relative_start_times[x])
                        ]
            self._sequence_thread.start()

    @property
    def paused(self) -> bool:
        """whether the sequence is paused"""
        return not self._sequence_pause.is_set()

    def pause_sequence(self):
        """Pauses the sequence (actions will not be automatically triggered until resumed)"""
        self._sequence_pause.clear()

    def kill_sequence(self):
        """This method will kill the sequence and perform cleanup. You will not be able to restart the thread
        after calling this method"""
        self._killswitch.set()  # activates killswitch
        self._sequence_pause.clear()  # clear pause if set (breaks out of wait section in loop)

    def join(self):
        """waits for defined time points in the scheduler to finish"""
        while self.remaining_time_points != 0:
            time.sleep(self.CYCLE_TIME)
        # if the last triggered time point is not complete, wait for completion
        last_timepoint = self.last_triggered_time_point
        if last_timepoint is not None and last_timepoint.complete is False:
            last_timepoint.wait_for_completion()

    def as_dict(self) -> dict:
        """Returns a key: value representation of the instance"""
        out = {
            'uuid': self.uuid,
            'sequence_name': self.sequence_name,
            'instance_actions': [action.name for action in self.sequence.action_sequence],
            'started': self.started,
            'created': self.created,
            'num_time_points': len(self.time_points),
            'sampled_time_points': self.triggered_time_points,
            'remaining_time_points': self.remaining_time_points,
        }
        if self.next_time_point is not None:
            out['next_time_point'] = self.next_time_point.time_delta
        if self._killswitch.is_set() is True:
            out['terminated'] = True
        return out

    def get_time_point_return(self,
                              key: str = None
                              ) -> dict:
        """
        Retrieves the returned values of the specified action time point.

        :param key: identifier key for the desired ActionTimePoint instance. If not specified, the last triggered
            time point will be retrieved.
        :return: method returns
        """
        if key is None:
            tp = self.last_triggered_time_point
        else:
            tp = ActionTimePoint.class_instance_by_id(key)
        return tp.action_returns

    def visualize_timepoints(self):
        raise NotImplementedError()
        pass  # todo plot a timeline of points for visualization

    def write_timepoints_to_csv(self,
                                file_name: str = None,
                                overwrite: bool = True,
                                ):
        """
        Writes the time point data to the specified csv file. If no filename is passed, the time points will be saved to
        "YYYY-MM-DD scheduler output.csv".

        :param file_name: file name to write to
        :param overwrite: whether to overwrite the filename if it already exists
        """
        # todo update to support new timepoint structure
        raise NotImplementedError()
        if file_name is None:
            file_name = f'{str(datetime.datetime.fromtimestamp(self._start_time).date())} scheduler output'
            if os.path.isfile(f'{file_name}.csv') and overwrite is False:
                i = 1
                while os.path.isfile(f'{file_name} - {i}.csv'):
                    i += 1
                file_name = f'{file_name} - {i}.csv'
            else:
                file_name += '.csv'
        self.logger.info(f'Writing scheduler data to {file_name}')
        with open(file_name, 'wt') as f:
            f.write(
                'Target Time (s),'
                'Sampled At (s),'
                'Timestamp,'
                # 'Metadata,'  # todo save this to file
                '\n'
            )
            if self.started is True:
                f.write(f'0.,0.,{str(datetime.datetime.fromtimestamp(self._start_time))}\n')
            # todo sort through time points and save negative timepoints before zero
            for tp in self.time_points:
                f.write(f'{tp.time_delta},')
                if tp.sampled_time is not None:
                    f.write(
                        f'{tp.sampled_time_relative if tp.sequence_time_started is not None else ""},'
                        f'{str(datetime.datetime.fromtimestamp(tp.sampled_time))},'
                    )
                f.write('\n')




import time
import unittest
from hein_control import SamplingScheduler, ConfiguredAction


# logging.basicConfig(level=logging.INFO)


@ConfiguredAction
def sleepy(t):
    time.sleep(t)


@ConfiguredAction
def slow_step():
    sleepy(0.1)


@ConfiguredAction
def fast_step():
    sleepy(0.01)


class TestScheduler(unittest.TestCase):
    def test_timepoint_addition(self):
        """tests timepoint addition to a scheduler instance"""
        sched = SamplingScheduler('slow_step', fast_step)
        sched.insert_time_point(0.05, number_of_points=4)
        self.assertEqual(len(sched.time_points), 4)
        # todo
        #   - test relative insertion
        #   - test list insertion
        #   - test list pairs insertion

    def test_execution(self):
        """tests execution of the scheduler"""
        sched = SamplingScheduler(slow_step, 'fast_step')
        sched.insert_time_point(0.5, number_of_points=4)
        self.assertFalse(sched.started)
        sched.start_sequence()
        self.assertTrue(sched.started)
        self.assertFalse(sched.paused)
        sched.pause_sequence()
        self.assertTrue(sched.paused)
        sched.start_sequence()
        sched.join()
        self.assertFalse(sched.busy)

    def test_triggering(self):
        """tests manual triggering of a scheduler"""
        sched = SamplingScheduler('fast_step')
        sched.trigger()
        self.assertIsNotNone(sched.last_triggered_time_point)
        sched.wait_for_time_point_completion()

    def test_retrieval(self):
        """tests value retrieval from a scheduler"""
        sched = SamplingScheduler('fast_step')
        sched.insert_time_point(0.)
        self.assertIsNone(sched.last_triggered_time_point)
        sched.start_sequence()
        sched.join()
        self.assertIsNotNone(sched.last_triggered_time_point)
        sched.as_dict()

    def test_too_fast(self):
        """tests what happens when time points are executed too fast"""
        sched = SamplingScheduler('slow_step', fast_step)
        sched.insert_time_point(0.001, number_of_points=4)
        sched.start_sequence()
        sched.trigger()
        sched.join()

    def test_duplicate_naming(self):
        """tests catch and acceptance of duplicate configuration names"""
        prev_name = ''
        for i in range(10):
            sched = SamplingScheduler(
                sequence_name='convenience name'
            )
            self.assertNotEqual(prev_name, sched.sequence.name)
            prev_name = sched.sequence_name

    # def test_addition_actions(self):  # todo
    #     pass

    # def test_action_override(self):  # todo
    #     pass



"""
Tools for executing a predefined (linear) series of actions.
"""

import logging
import datetime

from typing import Dict, Union, Iterable

from ..action.operators import execute_sequence_actions
from ..action.configured import ConfiguredAction, ThreadedTrackedAction, TrackedActionList, ConfiguredActionList, ensure_triggered

logger = logging.getLogger(__name__)


class ConfiguredSequence(ConfiguredAction):
    def __init__(self,
                 *actions: ConfiguredAction,
                 sequence_name: str = None,
                 next_action: Union[str, 'ConfiguredAction'] = None,
                 **kwargs,
                 ):
        """
        A class for configuring the execution of a sequence of actions

        :param actions: actions to execute in sequence
        :param sequence_name: name for the sequence
        :param next_action: next action to perform after completion of the sequence
        :param kwargs: kwargs for the actions
        """
        ConfiguredAction.__init__(
            self,
            action=execute_sequence_actions,
            configuration_name=sequence_name,  # todo add attribute for sequence name
            next_action=next_action,
            **kwargs,
        )
        self._sequence: ConfiguredActionList = ConfiguredActionList(*actions)

    def __str__(self):
        return f'{self.__class__.__name__} {self.name}'

    @property
    def action_sequence(self) -> ConfiguredActionList:
        """sequence of actions for the configuration"""
        return self._sequence

    @action_sequence.setter
    def action_sequence(self, value: Iterable):
        self._sequence = ConfiguredActionList(*value)

    def get_tracked_from_config(self, **kwargs) -> "ExecutionSequence":
        """
        Gets an execution sequence instance based on the configured sequence.

        :param kwargs: additional kwargs that differ from the instance configuration
        :return: execution sequence
        """
        return ExecutionSequence(configuration=self, **kwargs)

    def duplicate_configuration(self, new_name: str = None) -> 'ConfiguredSequence':
        """
        Duplicates the configuration of the sequence

        :param new_name: optional name for the new sequence
        :return:
        """
        if new_name is None:
            new_name = self._generate_unique_name(self.action)
        return self.__class__(
            *self.action_sequence,
            sequence_name=new_name,
            next_action=self.next_action,
            **self.kwargs,
        )


class ExecutionSequence(ThreadedTrackedAction):
    UUID_PREFIX = 'seq-'

    def __init__(self,
                 configuration: Union[str, ConfiguredSequence],
                 **kwargs,
                 ):
        """
        Executor class for a sequence of actions.

        :param actions: actions to execute when called (base actions or configured actions)
        """
        if isinstance(configuration, ConfiguredSequence) is False:
            configuration = ConfiguredSequence.class_instance_by_id(configuration)
        self._configuration: ConfiguredSequence = None  # update type hint
        ThreadedTrackedAction.__init__(
            self,
            configuration=configuration,
            **kwargs,
        )
        # todo disable args at the appropriate level (somewhere it breaks, figure out where that is)

    def __str__(self):
        # todo update to better represent instance
        return super(ExecutionSequence, self).__str__()

    def __call__(self, **kwargs):
        logger.debug(f'{self} sequence execution starting')
        # convert configured actions to tracked actions
        actions = TrackedActionList(*self.configuration.action_sequence)
        super(ExecutionSequence, self).__call__(
            *actions,
            **kwargs,
        )
        logger.debug(f'{self} sequence execution complete')

    @property
    def configuration(self) -> ConfiguredSequence:
        """the configured sequence associated with the instance"""
        return self._configuration

    @configuration.setter
    def configuration(self, value: ConfiguredSequence):
        if isinstance(value, ConfiguredSequence) is False:
            raise TypeError(f'a ConfiguredSequence must be used as a configuration for {self.__class__.__name__}')
        self._configuration = value
        self.update_args(*self._configuration.args)
        self.update_kwargs(*self._configuration.kwargs)

    def copy_configuration(self) -> "ExecutionSequence":
        """
        Copies the configuration of a ExecutionSequence instance.

        :return: ExecutionSequence instance with identical configuration
        """
        return self.__class__(
            self.configuration,
            **self.kwargs,
        )

    @property
    @ensure_triggered
    def action_returns(self) -> dict:
        """retrieves the returned values from the actions. Only actions which have completed will have return values"""
        return {
            action.configuration.name: action.action_return for action in self.args
        }

    @property
    @ensure_triggered
    def actions_status(self) -> dict:
        """status of the actions of the instance"""
        # todo consider passing through configured status
        return {action.configuration.name: action.status for action in self.args}

    @property
    @ensure_triggered
    def action_start_times(self) -> dict:
        """times when the prescribed actions were started"""
        if self.triggered is True:
            return {
                action.configuration.name: action.time_started for action in self.args
            }

    @property
    @ensure_triggered
    def action_completion_times(self) -> dict:
        """times when the prescribed actions were completed"""
        if self.triggered is True:
            return {
                action.configuration.name: action.time_completed for action in self.args
            }

    @property
    @ensure_triggered
    def action_durations(self) -> dict:
        """execution durations for the prescribed actions"""
        if self.triggered is True:
            return {
                action.configuration.name: action.action_duration for action in self.args
            }

    @property
    @ensure_triggered
    def started_timestamps(self) -> Dict[str, datetime.datetime]:
        """timestamps for when actions were started"""
        if self.triggered is True:
            return {
                action.configuration.name: action.started_timestamp for action in self.args
            }

    @classmethod
    def build_from_first_action(cls, first_action: ConfiguredAction) -> "ExecutionSequence":
        """
        Builds an execution sequence from a first action and its linked action

        :param first_action: first action to reference
        :return: execution sequence
        """
        raise NotImplementedError  # todo
